import math
import os
import pickle
from datetime import datetime
from re import A
from typing import List

import mlflow
import numpy as np
import pandas as pd
#import pygmo as pg
from mlflow.tracking import MlflowClient
from sklearn.metrics import *
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler
from tqdm import tqdm

from autooc.algorithm.parameters import params, set_params
from autooc.fitness.evaluation import evaluate_fitness
from autooc.operators.initialisation import initialisation
from autooc.stats.stats import get_stats, stats
from autooc.utilities.algorithm.general import check_python_version
from autooc.utilities.algorithm.initialise_run import initialise_run_params
from autooc.utilities.algorithm.NSGA2 import compute_pareto_metrics

check_python_version()


class AutoOC(object):
    def __init__(self, anomaly_class: str, normal_class: str, multiobjective: bool = False, algorithm: str = "all", performance_metric: str = "num_params", multicore: bool = False):
        #self.params = params.copy()
        self.params = params
        self.population = []
        self.population_dict = {}
        self.stats = {}
        self.fitted = False
        self.anomaly_class = anomaly_class
        self.normal_class = normal_class
        self.multiobjective = multiobjective
        self.algorithm = algorithm
        self.performance_metric = performance_metric
        self.multicore = multicore
        self.leaderboard = None

        init_params = {
            "ANOMALY_CLASS": self.anomaly_class,
            "NORMAL_CLASS": self.normal_class,
            "ALGORITHM": self.algorithm,
            "MULTICORE": self.multicore,
            "FITNESS_FUNCTION": f"supervised_learning.classification, {self.performance_metric}" if self.multiobjective else "supervised_learning.classification"
        }

        params.update(init_params)
        self.params.update(init_params)

    def fit(
        self,
        X: pd.DataFrame,
        X_val: pd.DataFrame,
        y_val: pd.DataFrame = None,
        pop: int = 100,
        gen: int = 100,
        early_stopping_rounds: int = None,
        early_stopping_tolerance: float = 0.01,
        always_at_hand: bool = False,
        results_path: str = "./",
        mlflow_tracking_uri: str = "./",
        mlflow_experiment_name: str = "mlflow_experiment",
        mlflow_run_name: str = f"run",
        **extra_params
    ) -> None:

        ALGORITHMS = {
            "autoencoder": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "autoencoders_v4.pybnf"),
                "metric_supervised": "auc_autoencoder",
                "metric_unsupervised": "reconstruction_error"
            },
            "iforest": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "iforest.pybnf"),
                "metric_supervised": "auc_sklearn",
                "metric_unsupervised": "anomaly_score_iforest"
            },
            "svm": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "svm.pybnf"),
                "metric_supervised": "auc_sklearn",
                "metric_unsupervised": "anomaly_score_svm"
            },
            "lof": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "lof.pybnf"),
                "metric_supervised": "auc_sklearn",
                "metric_unsupervised": "anomaly_score_lof"
            },
            "vae": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "vae.pybnf"),
                "metric_supervised": "auc_autoencoder",
                "metric_unsupervised": "reconstruction_error"
            },
            "nas": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "nas.pybnf"),
                "metric_supervised": "auc_autoencoder",
                "metric_unsupervised": "reconstruction_error"
            },
            "all": {
                "grammar": os.path.join(os.path.dirname(os.path.abspath(__file__)), "grammars", "multi_algo.pybnf"),
                "metric_supervised": "auc_autoencoder",
                "metric_unsupervised": "reconstruction_error"
            }
        }

        fit_params = {
            "X_train": X,
            "X_train_len": len(X),
            "X_val": X_val,
            "y_val": y_val,
            "POPULATION_SIZE": pop,
            "GENERATIONS": gen,
            "EXPERIMENT_NAME": mlflow_experiment_name,
            "TYPE": "unsupervised" if y_val is None else "supervised",
            # "FITNESS_FUNCTION": f"supervised_learning.classification, {self.performance_metric}" if self.multiobjective else "supervised_learning.classification",
            "RESULTS_PATH": os.path.abspath(results_path),
            "ALWAYS_AT_HAND": always_at_hand,
        }

        if self.algorithm in ALGORITHMS:
            fit_params["GRAMMAR_FILE"] = ALGORITHMS[self.algorithm]["grammar"]
            fit_params["ERROR_METRIC"] = ALGORITHMS[
                self.algorithm][f"metric_{fit_params['TYPE']}"]
        else:
            fit_params["GRAMMAR_FILE"] = ALGORITHMS["all"]["grammar"]
            fit_params["ERROR_METRIC"] = ALGORITHMS["all"][f"metric_{fit_params['TYPE']}"]

        if self.multiobjective:
            fit_params["REPLACEMENT"] = "nsga2_replacement"
            fit_params["SELECTION"] = "nsga2_selection"
        else:
            fit_params["REPLACEMENT"] = "generational"
            fit_params["SELECTION"] = "tournament"

        params.update(fit_params)
        self.params.update(fit_params)

        param_list = list_params(**extra_params)
        set_params(param_list)

        #self.params = params

        # initialize run
        initialise_run_params(True)

        # initialise population
        self.population = initialisation(self.params["POPULATION_SIZE"])

        # Evaluate initial population
        self.population = evaluate_fitness(self.population)

        stats["gen"] = 0

        # Generate statistics for run so far
        get_stats(self.population)

        #mlflow = get_mlflow(self.params["EXPERIMENT_NAME"])
        mlflow_params = {
            "run_name": mlflow_run_name,
            "tracking_uri": mlflow_tracking_uri,
            "experiment_name": mlflow_experiment_name,
        }

        range_generations = tqdm(range(1, (self.params["GENERATIONS"] + 1)))

        self.population = self.evolve(
            self.params, range_generations, self.population, early_stopping_rounds, early_stopping_tolerance, mlflow_params)
        # get_stats(population, end=True)
        store_pop(self.population)

        #print(self.population)
        self.population_dict = {ind.id: ind for ind in self.population}

        self.stats = stats
        #self.population = population
        self.fitted = True

    def reconstruct(self, X_test: pd.DataFrame, mode: str = "all"):
        if mode == "all":
            reconstrutions = [ind.reconstruct(X_test)
                              for ind in self.population]
        elif mode == "best":
            if self.multiobjective:
                best = min(self.population, key=lambda x: x.fitness[0])
            else:
                best = min(self.population, key=lambda x: x.fitness)
            reconstrutions = best.reconstruct(X_test)
        elif mode == "simplest":
            simplest = min(self.population, key=lambda x: x.fitness[1])
            reconstrutions = simplest.reconstruct(X_test)
        elif mode == "balanced":
            from math import log10

            min_y = log10(
                min(self.population, key=lambda x: x.fitness[1]).fitness[1])
            max_y = log10(
                max(self.population, key=lambda x: x.fitness[1]).fitness[1])
            # get individual with greater distance to point (0, 1)
            balanced = max(self.population,
                           key=lambda x: get_distance(x, min_y, max_y))
            reconstrutions = balanced.reconstuct(X_test)

        return reconstrutions

    def predict(self, X_test: pd.DataFrame, mode: str = "all", threshold="default", percentile=95) -> list:
        individuals = self.get_individuals(mode)
        if threshold == "default":
            return [ind.predict(X_test, self.anomaly_class, self.normal_class, ind.threshold) for ind in individuals]

        elif type(threshold) in [int, float]:
            return [ind.predict(X_test, self.anomaly_class, self.normal_class, threshold) for ind in individuals]

        else:
            thresholds = self.calculate_threshold(
                mode, threshold, percentile=percentile)
            return [individuals[i].predict(X_test, self.anomaly_class, self.normal_class, thresholds[i]) for i in range(len(individuals))]

    def evaluate(
        self,
        X_test: pd.DataFrame,
        y_test: pd.DataFrame,
        mode: str = "all",
        metric: str = "roc_auc",
        threshold="default",
        percentile=95
    ) -> list:
        metrics = {
            "roc_auc": roc_auc_score,
            "accuracy": accuracy_score,
            "precision": precision_score,
            "recall": recall_score,
            "f1": f1_score,
        }

        preds = self.predict(X_test, mode, threshold, percentile)

        if isinstance(self.anomaly_class, str):
            preds = [(lambda x: x.replace(self.anomaly_class, '1'))(x)
                     for x in preds]
            preds = [(lambda x: x.replace(self.normal_class, '0'))(x)
                     for x in preds]
            #print(preds[0])
            preds = [(lambda x: x.astype(int))(x) for x in preds]

            y_test = (lambda x: x.replace(self.anomaly_class, '1'))(y_test)
            y_test = (lambda x: x.replace(self.normal_class, '0'))(y_test)
            #print(y_test)
            y_test = (lambda x: x.astype(int))(y_test)

        scores = [metrics[metric](y_test, pred) for pred in preds]

        return scores

    def evaluate_all(self, X_test: pd.DataFrame, y_test: pd.Series, metric: str = "roc_auc") -> List:
        """
        Evaluate all individuals in terms of AUC and Complexity,\
            based on data set (X_test, y_test)
        Only works for multiobjetive problems

        Parameters
        ----------
        X_test : pd.DataFrame
            The test input samples.
        y_test : pd.Series
            The test target values (class labels) as integers or strings.

        Returns
        -------
        List
            List containing [AUC, Complexity] for each Evolutionary Decision Tree in the population.
        """
        performances = self.evaluate(X_test, y_test, metric=metric, mode="all")
        # performances = [metrics[metric](y_test, ind.predict(X_test, self.anomaly_class, self.normal_class))
        #        for ind in self.population]

        complexities = [ind.fitness[1] for ind in self.population]
        ev = pd.DataFrame([performances, complexities]).T
        ev.columns = ['performance', 'complexity']
        ev2 = ev.groupby('performance').agg(
            {'complexity': min}).reset_index().values

        return [list(ev2[:, 0]), list(ev2[:, 1])]

    def load_example_data(self) -> np.array:
        # Download the dataset
        PATH_TO_DATA = (
            "http://storage.googleapis.com/download.tensorflow.org/data/ecg.csv"
        )
        data = pd.read_csv(PATH_TO_DATA, header=None).add_prefix("col_")

        # last column is the target
        # 0 = anomaly, 1 = normal
        TARGET = "col_140"

        features = data.drop(TARGET, axis=1)
        target = data[TARGET]

        X_train, X_test, y_train, y_test = train_test_split(
            features, target, test_size=0.2, stratify=target
        )

        train_index = y_train[y_train == 1].index
        train_data = X_train.loc[train_index]

        min_max_scaler = MinMaxScaler(feature_range=(0, 1))
        X_train_scaled = min_max_scaler.fit_transform(train_data.copy())
        X_train_scaled, X_val_scaled = train_test_split(
            X_train_scaled, test_size=0.25)
        X_test_scaled = min_max_scaler.transform(X_test.copy())

        return X_train_scaled, X_val_scaled, X_test_scaled, y_test

    def split_data(self, df: pd.DataFrame, target_col: str, normalize: bool = True) -> pd.DataFrame:
        features = df.drop(target_col, axis=1)
        target = df[target_col]

        X_train, X_test, y_train, y_test = train_test_split(
            features, target, test_size=0.2, stratify=target
        )

        train_index = y_train[y_train == self.normal_class].index
        train_data = X_train.loc[train_index]

        if normalize:
            min_max_scaler = MinMaxScaler(feature_range=(0, 1))
            X_train_scaled = min_max_scaler.fit_transform(train_data.copy())
            X_train_scaled, X_val_scaled = train_test_split(
                X_train_scaled, test_size=0.25)
            X_test_scaled = min_max_scaler.transform(X_test.copy())
            return X_train_scaled, X_val_scaled, X_test_scaled, y_test
        else:
            X_train, X_val = train_test_split(
                X_train, test_size=0.25)
            return X_train, X_val, X_test, y_test

    def calculate_threshold(self, mode: str = "all", method: str = "mean", percentile=95):
        individuals = self.get_individuals(mode)
        if method == "mean":
            thresholds = [(np.mean(ind.losses) + np.std(ind.losses))
                          for ind in individuals]
        elif method == "percentile":
            thresholds = [np.percentile(ind.losses, percentile)
                          for ind in individuals]
        elif method == "max":
            thresholds = [np.max(ind.losses) for ind in individuals]

        return thresholds

    def get_individuals(self, mode: str = "all") -> list:
        if mode == "all":
            individuals = self.population
        elif mode == "best":
            if self.params["ERROR_METRIC"].maximise:
                individuals = [max(self.population, key=lambda x: x.fitness[0])] if self.multiobjective else [
                    max(self.population, key=lambda x: x.fitness)]
            else:
                individuals = [min(self.population, key=lambda x: x.fitness[0])] if self.multiobjective else [
                    min(self.population, key=lambda x: x.fitness)]
        elif mode == "pareto":
            individuals = self.get_pareto_front()
        elif mode == "simplest":
            individuals = [min(self.population, key=lambda x: x.fitness[1])]
        return individuals

    def get_pareto_front(self) -> list:
        return compute_pareto_metrics(self.population).fronts[0]

    def get_statistics(self) -> dict:
        return self.stats

    def get_performance_metric(self) -> str:
        return self.performance_metric

    def get_population(self) -> dict:
        return self.population

    def get_population_dict(self) -> dict:
        return self.population_dict

    def get_params(self) -> dict:
        return self.params

    def get_classes(self) -> dict:
        return {
            "anomaly_class": self.anomaly_class,
            "normal_class": self.normal_class
        }

    def get_default_thresholds(self) -> list:
        return [ind.threshold for ind in self.population]

    def get_individual(self, id):
        return self.population_dict[id]

    def get_leaderboard(self, orderby="generation") -> dict:
        if orderby == "generation":
            return self.leaderboard
        elif orderby == "performance":
            return self.leaderboard.sort_values(by="reconstruction_error", ascending=True)
        elif orderby == "complexity":
            return self.leaderboard.sort_values(by=self.performance_metric, ascending=True)

    def set_classes(self, anomaly_class, normal_class) -> None:
        self.anomaly_class = anomaly_class
        self.normal_class = normal_class

    def evolve(
        self,
        params: dict,
        range_generations: range,
        population: list,
        early_stopping_rounds: int,
        early_stopping_tolerance: float,
        mlflow_params: dict = None,
    ) -> List:

        leaderboard_list = []

        # config mlflow experiment
        mlflow.set_tracking_uri(os.path.join(
            os.path.abspath(mlflow_params["tracking_uri"]), "mlruns"))
        mlflow.set_registry_uri(os.path.join(
            os.path.abspath(mlflow_params["tracking_uri"]), "mlruns"))
        mlflow.set_experiment(mlflow_params["experiment_name"])

        # trigger to end current run faster
        end = False
        round_count = 0  # rounds without improvement

        if self.params["ERROR_METRIC"].maximise:
            best_performance = 0
        else:
            best_performance = np.inf

        with mlflow.start_run(run_name=mlflow_params["run_name"]) as run:
            for key, val in params.items():
                if key in [
                    "X_train",
                    "y_train",
                    "X_train_len",
                    "X_val",
                    "y_val",
                    "X_test",
                    "y_test",
                    "POOL",
                    "BNF_GRAMMAR",
                    "SEED_INDIVIDUALS",
                ]:
                    continue
                else:
                    mlflow.log_param(key, val)

            for generation in range_generations:
                best_performance_before_generation = best_performance
                stats["gen"] = generation

                population = params["STEP"](population)

                i = 1
                for ind in population:
                    leaderboard_dict = {
                        "ID": ind.id,
                        "generation": generation,
                        "individual": i
                    }

                    if self.multiobjective:
                        leaderboard_dict[self.params["ERROR_METRIC"].__name__] = ind.fitness[0],
                        leaderboard_dict[self.performance_metric] = ind.fitness[1]
                    else:
                        leaderboard_dict[self.params["ERROR_METRIC"].__name__] = ind.fitness
                        #print("INDIVIDUAL FITNESS: \n\n\n\n\n\n\n", ind.fitness)
                        if self.params["ERROR_METRIC"].maximise:
                            if ind.fitness > best_performance:
                                best_performance = ind.fitness
                        else:
                            if ind.fitness < best_performance:
                                best_performance = ind.fitness

                    leaderboard_list.append(leaderboard_dict)
                    i += 1

                if self.multiobjective:
                    performances = [ind.fitness[0] for ind in population]
                    complexities = [ind.fitness[1] for ind in population]
                    generation_metrics = {
                        "1st ind PERFORMANCE": min(performances),
                        "1st ind COMPLEXITY": max(complexities),
                        "last ind PERFORMANCE": max(performances),
                        "last ind COMPLEXITY": min(complexities),
                        "mean PERFORMANCE": np.mean(performances),
                        "mean COMPLEXITY": np.mean(complexities),
                    }
                else:
                    performances = [ind.fitness for ind in population]
                    generation_metrics = {
                        "1st ind PERFORMANCE": min(performances),
                        "last ind PERFORMANCE": max(performances),
                        "mean PERFORMANCE": np.mean(performances)
                    }

                mlflow.log_metrics(
                    metrics=generation_metrics,
                    step=generation,
                )

                if self.multiobjective:
                    # get pareto front from the current generation
                    pareto_individuals = compute_pareto_metrics(
                        population).fronts[0]
                    if self.params["ERROR_METRIC"].maximise:
                        pareto_points = [[-ind.fitness[0], ind.fitness[1]]
                                         for ind in pareto_individuals]
                    else:
                        pareto_points = [[ind.fitness[0], ind.fitness[1]]
                                         for ind in pareto_individuals]

                    #hypervolume = pg.hypervolume(pareto_points)
                    #ref_value = hypervolume.refpoint()
                    #hypervolume_value = hypervolume.compute(ref_value)
                    #print("HYPERVOLUME: \n\n\n\n\n\n\n\n\n\n\n", hypervolume_value)

                    # if hypervolume_value > best_performance:
                    #    best_performance = hypervolume_value

                    if (best_performance + early_stopping_tolerance) > best_performance_before_generation:
                        round_count = 0
                    else:
                        round_count += 1
                else:
                    if self.params["ERROR_METRIC"].maximise:
                        if (best_performance + early_stopping_tolerance) > best_performance_before_generation:
                            round_count = 0
                        else:
                            round_count += 1
                    else:
                        if (best_performance - early_stopping_tolerance) < best_performance_before_generation:
                            round_count = 0
                            best_performance_before_generation = best_performance
                        else:
                            round_count += 1

                #print(f"ROUND COUNT \n\n\n\n\n\n\:{round_count}")

                if early_stopping_rounds is not None and round_count >= early_stopping_rounds:
                    end = True
                    #print("EARLY STOPPING\n\n\n\n\n\n\n")

                if generation == params["GENERATIONS"] or end:
                    i = 0
                    for ind in population:
                        if ind.model.__class__.__name__ == "Sequential":
                            #mlflow.keras.log_model(ind.model, f"model_{i}")
                            pass
                        elif ind.model.__class__.__name__ in ["OneClassSVM", "IsolationForest"]:
                            #mlflow.sklearn.log_model(ind.model, f"model_{i}")
                            pass
                        i += 1

                    if end:
                        break

            get_stats(population, end=True)
            artifacts_dir = params["FILE_PATH"]
            mlflow.log_artifacts(artifacts_dir, artifact_path="images")

            # save leaderboard
            leaderboard = pd.DataFrame.from_dict(leaderboard_list)
            leaderboard.to_csv(params["FILE_PATH"] +
                               "/leaderboard.csv", index=False)

            mlflow.log_artifact(params["FILE_PATH"] + "/leaderboard.csv")
            self.leaderboard = leaderboard.copy()

        # mlflow.end_run()

        return population


def list_params(**extra_params: dict) -> List:
    """
    Internal function to list execution parameters.
    For advanced users only.
    Parameters
    ----------
    **extra_params : dict
        Extra parameters. For details, please check: https://github.com/PonyGE/PonyGE2/wiki/Evolutionary-Parameters.
    Returns
    -------
    param_list : List
        List of parameters.
    """
    param_list = []
    for key, val in extra_params.items():
        if val == "True":
            param_list.append("--" + key)
        elif val in ["False", ""]:
            continue
        else:
            param_list.append("--{0}={1}".format(key, str(val)))
    return param_list


def store_pop(population: List):
    """
    Stores the evolved population. For advanced users only.
    Parameters
    ----------
    population : List
        The population to be stored
    Returns
    -------
    None.
    """

    # SEEDS_PATH = path.join('results', params["FILE_PATH"], 'seeds')
    # makedirs(path.join(getcwd(), SEEDS_PATH, params['TARGET_SEED_FOLDER']),
    #         exist_ok=True)
    SEEDS_PATH = os.path.join(params["FILE_PATH"], "seeds")
    os.makedirs(SEEDS_PATH, exist_ok=True)
    for cont, item in enumerate(population):
        if item.phenotype != None:
            # fname = path.join(SEEDS_PATH, params['TARGET_SEED_FOLDER'],
            #                  "{0}.txt".format(str(cont)))
            fname = os.path.join(
                SEEDS_PATH, params["TARGET_SEED_FOLDER"], "{0}.txt".format(
                    str(cont))
            )
            with open(fname, "w+", encoding="utf-8") as f:
                f.write("Phenotype:\n")
                f.write("%s\n" % item.phenotype)
                f.write("Genotype:\n")
                f.write("%s\n" % item.genome)
                f.write("Tree:\n")
                f.write("%s\n" % str(item.tree))
                f.write("Training fitness:\n")
                f.write("%s\n" % item.fitness)
                f.close()


def get_distance(ind, min_y, max_y):
    auc = ind.fitness[0] / -100  # auc (positive, from 0 to 1)
    comp = math.log10(ind.fitness[1])  # complexity
    # scale complexity to [0, 1]
    comp = (comp - min_y) / (max_y - min_y)
    # worst result: (0, 1)
    x = 0
    y = 1
    # get distance:
    dist = math.hypot(auc - x, comp - y)
    return dist



from os import path, pathsep

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from autooc.utilities.stats.trackers import first_pareto_list

matplotlib.use("Agg")

plt.rc("font", family="Times New Roman")


def save_pareto_fitness_plot():
    """
    Saves a plot of the current fitness for a pareto front.

    :return: Nothing
    """

    from autooc.algorithm.parameters import params

    # Initialise up figure instance.
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)

    # Set up iterator for color plotting.
    color = iter(plt.cm.jet(np.linspace(0, 1, len(first_pareto_list))))

    # Get labels for individual fitnesses.
    ffs = params["FITNESS_FUNCTION"].fitness_functions

    # Find the direction for step lines to "bend"
    step_dir = "pre" if ffs[0].maximise else "post"

    # Plot data.
    for i, gen in enumerate(first_pareto_list):
        c = next(color)
        ax1.step(
            gen[0], gen[1], linestyle="--", where=step_dir, color=c, lw=0.35, alpha=0.25
        )
        ax1.plot(gen[0], gen[1], "o", color=c, ms=1)

    # Set labels with class names.
    ax1.set_xlabel(ffs[0].__class__.__name__, fontsize=14)
    ax1.set_ylabel(ffs[1].__class__.__name__, fontsize=14)

    # Plot title and legend.
    plt.title("First pareto fronts by generation")

    # Set up colorbar instead of legend. Normalise axis to scale of data.
    sm = plt.cm.ScalarMappable(
        cmap="jet", norm=plt.Normalize(vmin=0, vmax=len(first_pareto_list) - 1)
    )

    # Fake up the array of the scalar mappable.
    sm._A = []

    # Plot the colorbar.
    cbar = plt.colorbar(sm, ticks=[0, len(first_pareto_list) - 1])

    # Set label of colorbar.
    # cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel("Generation", rotation=90)

    # Save plot and close.
    plt.savefig(path.join(params["FILE_PATH"], "fitness.pdf"))
    plt.close()


def save_plot_from_data(data, name):
    """
    Saves a plot of a given set of data.

    :param data: the data to be plotted
    :param name: the name of the data to be plotted.
    :return: Nothing.
    """

    from autooc.algorithm.parameters import params

    # Initialise up figure instance.
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)

    # Plot data.
    ax1.plot(data)

    # Set labels.
    ax1.set_ylabel(name, fontsize=14)
    ax1.set_xlabel("Generation", fontsize=14)

    # Plot title.
    plt.title(name)

    # Save plot and close.
    plt.savefig(path.join(params["FILE_PATH"], (name + ".pdf")))
    plt.close()


def save_plot_from_file(filename, stat_name):
    """
    Saves a plot of a given stat from the stats file.

    :param filename: a full specified path to a .csv stats file.
    :param stat_name: the stat of interest for plotting.
    :return: Nothing.
    """

    # Read in the data
    data = pd.read_csv(filename, sep="\t")
    try:
        stat = list(data[stat_name])
    except KeyError:
        s = (
            "utilities.stats.save_plots.save_plot_from_file\n"
            "Error: stat %s does not exist" % stat_name
        )
        raise Exception(s)

        # Set up the figure.
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)

    # Plot the data.
    ax1.plot(stat)

    # Plot title.
    plt.title(stat_name)

    # Get save path
    save_path = pathsep.join(filename.split(pathsep)[:-1])

    # Save plot and close.
    plt.savefig(path.join(save_path, (stat_name + ".pdf")))
    plt.close()


def save_box_plot(data, names, title):
    """
    Given an array of some data, and a list of names of that data, generate
    and save a box plot of that data.

    :param data: An array of some data to be plotted.
    :param names: A list of names of that data.
    :param title: The title of the plot.
    :return: Nothing
    """

    import matplotlib.pyplot as plt

    from autooc.algorithm.parameters import params

    plt.rc("font", family="Times New Roman")

    # Set up the figure.
    fig = plt.figure()
    ax1 = fig.add_subplot(1, 1, 1)

    # Plot tight layout.
    plt.tight_layout()

    # Plot the data.
    ax1.boxplot(np.transpose(data), 1)

    # Plot title.
    plt.title(title)

    # Generate list of numbers for plotting names.
    nums = list(range(len(data))[1:]) + [len(data)]

    # Plot names for each data point.
    plt.xticks(nums, names, rotation="vertical", fontsize=8)

    # Save plot.
    plt.savefig(path.join(params["FILE_PATH"], (title + ".pdf")))

    # Close plot.
    plt.close()



from autooc.algorithm.parameters import params
from autooc.fitness.evaluation import evaluate_fitness
from autooc.operators.crossover import crossover_inds
from autooc.operators.mutation import mutation
from autooc.operators.selection import selection
from autooc.utilities.algorithm.NSGA2 import compute_pareto_metrics


def replacement(new_pop, old_pop):
    """
    Given a new population and an old population, performs replacement using
    specified replacement operator.

    :param new_pop: Newly generated population (after selection, variation &
    evaluation).
    :param old_pop: The previous generation population.
    :return: Replaced population.
    """
    return params["REPLACEMENT"](new_pop, old_pop)


def generational(new_pop, old_pop):
    """
    Replaces the old population with the new population. The ELITE_SIZE best
    individuals from the previous population are appended to new pop regardless
    of whether or not they are better than the worst individuals in new pop.

    :param new_pop: The new population (e.g. after selection, variation, &
    evaluation).
    :param old_pop: The previous generation population, from which elites
    are taken.
    :return: The 'POPULATION_SIZE' new population with elites.
    """

    # Sort both populations.
    old_pop.sort(reverse=True)
    new_pop.sort(reverse=True)

    # Append the best ELITE_SIZE individuals from the old population to the
    # new population.
    for ind in old_pop[: params["ELITE_SIZE"]]:
        new_pop.insert(0, ind)

    # Return the top POPULATION_SIZE individuals of the new pop, including
    # elites.
    return new_pop[: params["POPULATION_SIZE"]]


def steady_state(individuals):
    """
    Runs a single generation of the evolutionary algorithm process,
    using steady state replacement:
        Selection
        Variation
        Evaluation
        Replacement

    Steady state replacement uses the Genitor model (Whitley, 1989) whereby
    new individuals directly replace the worst individuals in the population
    regardless of whether or not the new individuals are fitter than those
    they replace. Note that traditional GP crossover generates only 1 child,
    whereas linear GE crossover (and thus all crossover functions used in
    PonyGE) generates 2 children from 2 parents. Thus, we use a deletion
    strategy of 2.

    :param individuals: The current generation, upon which a single
    evolutionary generation will be imposed.
    :return: The next generation of the population.
    """

    # Initialise counter for new individuals.
    ind_counter = 0

    while ind_counter < params["POPULATION_SIZE"]:

        # Select parents from the original population.
        parents = selection(individuals)

        # Perform crossover on selected parents.
        cross_pop = crossover_inds(parents[0], parents[1])

        if cross_pop is None:
            # Crossover failed.
            pass

        else:
            # Mutate the new population.
            new_pop = mutation(cross_pop)

            # Evaluate the fitness of the new population.
            new_pop = evaluate_fitness(new_pop)

            # Sort the original population
            individuals.sort(reverse=True)

            # Combine both populations
            total_pop = individuals[: -len(new_pop)] + new_pop

            # Increment the ind counter
            ind_counter += params["GENERATION_SIZE"]

    # Return the combined population.
    return total_pop


def nsga2_replacement(new_pop, old_pop):
    """
    Replaces the old population with the new population using NSGA-II
    replacement. Both new and old populations are combined, pareto fronts
    and crowding distance are calculated, and the replacement population is
    computed based on crowding distance per pareto front.

    :param new_pop: The new population (e.g. after selection, variation, &
                    evaluation).
    :param old_pop: The previous generation population.
    :return: The 'POPULATION_SIZE' new population.
    """

    # Combine both populations (R_t = P_t union Q_t)
    new_pop.extend(old_pop)

    # Compute the pareto fronts and crowding distance
    pareto = compute_pareto_metrics(new_pop)

    # Size of the new population
    pop_size = params["POPULATION_SIZE"]

    # New population to replace the last one
    temp_pop, i = [], 0

    while len(temp_pop) < pop_size:
        # Populate the replacement population

        if len(pareto.fronts[i]) <= pop_size - len(temp_pop):
            temp_pop.extend(pareto.fronts[i])

        else:
            # Sort the current pareto front with respect to crowding distance.
            pareto.fronts[i] = sorted(
                pareto.fronts[i], key=lambda item: pareto.crowding_distance[item]
            )

            # Get number of individuals to add in temp to achieve the pop_size
            diff_size = pop_size - len(temp_pop)

            # Extend the replacement population
            temp_pop.extend(pareto.fronts[i][:diff_size])

        # Increment counter.
        i += 1

    return temp_pop


# Set attributes for all operators to define multi-objective operators.
nsga2_replacement.multi_objective = True




# read the contents of your README file
from pathlib import Path

import setuptools

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setuptools.setup(
    name="autooc",
    version="0.0.13.8",
    author_email="luis_ferreira223@hotmail.com",
    author="Luís Ferreira",
    description="AutoOC: Automated Machine Learning (AutoML) library for One-Class Learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    url="https://github.com/luisferreira97/AutoOC",
    License="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    py_modules=["autooc"],
    python_requires='>=3.6, <3.9',
    keywords=['automl', 'machine learning', 'one-class learning',
              'one-class classification', 'autoencoder', 'isolation forest', 'one-class svm'],
    include_package_data=True,
    #package_dir={"": "src/autooc"},
    install_requires=[
        'keras==2.6.0',
        'matplotlib==3.4.1',
        'mlflow==1.15.0',
        'pandas==1.2.4',
        'pydot==1.4.2',
        'numpy==1.19.2',
        'scikit-learn==1.0',
        'tensorflow==2.6.0',
        'tensorflow-estimator==2.6.0',
        'tqdm==4.60.0'
    ]
)


from autooc.algorithm.parameters import params
from autooc.fitness.supervised_learning.supervised_learning import \
    supervised_learning
from autooc.utilities.fitness.error_metric import rmse


class regression(supervised_learning):
    """Fitness function for regression. We just slightly specialise the
    function for supervised_learning."""

    def __init__(self):
        # Initialise base fitness function class.
        super().__init__()

        # Set error metric if it's not set already.
        if params["ERROR_METRIC"] is None:
            params["ERROR_METRIC"] = rmse

        self.maximise = params["ERROR_METRIC"].maximise



import itertools
import random

import numpy as np

from autooc.algorithm.parameters import params
from autooc.fitness.supervised_learning.regression import regression
from autooc.utilities.fitness.error_metric import rmse


class regression_random_polynomial(regression):
    """Fitness function for regression of random polynomials. We
    specialise the regression fitness function.

    The user must pass in the degree, n_vars, and n_samples. A random
    polynomial of the given degree and number of variables is
    generated (ie random coefficients are generated). It is evaluated
    at n_samples, randomly generated. That gives a target dataset for
    training. This is accomplished with, eg, --extra_parameters 5 1
    20, giving degree 5, 1 variable, 20 fitness cases.

    If the user requires test on unseen data, the polynomial is
    evaluated at a new sample of points, the same size. That is
    accomplished with --dataset_test Dummy (the "Dummy" argument would
    usually be a filename, but we do not attempt to read a file, but
    just take it as a switch to generate the separate test set).

    An example command-line is then:

    python ponyge.py --generations 10 --population 10 --fitness
    supervised_learning.regression_random_polynomial
    --extra_parameters 5 1 20 --grammar
    supervised_learning/supervised_learning.bnf --dataset_test Dummy

    """

    def __init__(self):
        # Don't call super().__init__() because it reads the training
        # (and test) data from files. We'll do everything else it
        # would have done.

        degree, n_vars, n_samples = map(int, params["EXTRA_PARAMETERS"])

        # may be needed if the grammar uses GE_RANGE:dataset_n_vars
        self.n_vars = n_vars

        # Set error metric if it's not set already.
        if params["ERROR_METRIC"] is None:
            params["ERROR_METRIC"] = rmse

        self.maximise = params["ERROR_METRIC"].maximise

        # create a random polynomial
        p = Polynomial.from_random(degree, n_vars)

        # generate a set of fitness cases for training
        self.training_in = np.random.random((n_vars, n_samples))
        self.training_exp = p.eval(self.training_in)

        # if we want a separate test set, generate a set of fitness
        # cases for it
        if params["DATASET_TEST"]:
            self.training_test = True
            self.test_in = np.random.random((n_vars, n_samples))
            self.test_exp = p.eval(self.training_in)


class Polynomial:

    """A polynomial of a given degree and a given number of variables,
    with one coefficient for each term."""

    def __init__(self, degree, n_vars, coefs):
        """Constructor for the case where we already have coefficients."""
        self.degree = degree
        self.n_vars = n_vars
        self.coefs = coefs

    @classmethod
    def from_random(cls, degree, n_vars):
        """Constructor for the case where we want random coefficients."""
        coefs = [2 * (random.random() - 0.5)
                 for i in Polynomial.terms(degree, n_vars)]
        p = Polynomial(degree, n_vars, coefs)
        return p

    @staticmethod
    def terms(degree, n_vars):
        """Generator for the possible terms of a polynomial
        given the degree and number of variables."""
        for pows in itertools.product(range(degree + 1), repeat=n_vars):
            if sum(pows) <= degree:
                yield pows

    def eval(self, x):
        """Evaluate the polynomial at a set of points x."""
        assert x.shape[0] == self.n_vars
        result = np.zeros(x.shape[1])  # same length as a column of x
        for coef, pows in zip(self.coefs, self.terms(self.degree, self.n_vars)):
            tmp = np.ones(x.shape[1])
            for (xi, pow) in zip(x, pows):
                tmp *= xi ** pow
            tmp *= coef
            result += tmp
        return result

    def __str__(self):
        """Pretty-print a polynomial, rounding the coefficients."""

        def s(pows):
            if sum(pows):
                return "*" + "*".join(
                    "x[%d]**%d" % (i, powi) for i, powi in enumerate(pows) if powi > 0
                )
            else:
                return ""  # this term is a const so the coef on its own is enough

        return " + ".join(
            "%.3f%s" % (coef, s(pows))
            for (coef, pows) in zip(
                self.coefs, Polynomial.terms(self.degree, self.n_vars)
            )
        )



import importlib
from datetime import datetime
from os import getpid
from random import seed
from socket import gethostname
from time import time

from autooc.algorithm.parameters import params
from autooc.utilities.stats import trackers
from autooc.utilities.stats.file_io import generate_folders_and_files


def initialise_run_params(create_files):
    """
    Initialises all lists and trackers. Generates save folders and initial
    parameter files if debugging is not active.

    :return: Nothing
    """

    start = datetime.now()
    trackers.time_list.append(time())

    # Set random seed
    if params["RANDOM_SEED"] is None:
        params["RANDOM_SEED"] = int(start.microsecond)
    seed(params["RANDOM_SEED"])

    # Generate a time stamp for use with folder and file names.
    hms = "%02d%02d%02d" % (start.hour, start.minute, start.second)
    params["TIME_STAMP"] = "_".join(
        [
            gethostname(),
            str(start.year)[2:],
            str(start.month),
            str(start.day),
            hms,
            str(start.microsecond),
            str(getpid()),
            str(params["RANDOM_SEED"]),
        ]
    )
    if not params["SILENT"]:
        print("\nStart:\t", start, "\n")

    # Generate save folders and files
    if params["DEBUG"]:
        print("Seed:\t", params["RANDOM_SEED"], "\n")
    elif create_files:
        generate_folders_and_files()


def set_param_imports():
    """
    This function makes the command line experience easier for users. When
    specifying operators listed in the lists below, users do not need to
    specify the full file path to the functions themselves. Users can simply
    specify a single word, e.g.

        "--mutation subtree"

    Using the special_ops dictionary for example, this will default to
    "operators.mutation.subtree. Executes the correct imports for specified
    modules and then saves the correct parameters in the params dictionary.
    Users can still specify the full direct path to the operators if they so
    desire, allowing them to create new operators and save them wherever
    they like.

    Sets the fitness function for a problem automatically. Fitness functions
    must be stored in fitness. Fitness functions must be classes, where the
    class name matches the file name.

    :return: Nothing.
    """

    # For these ops we let the param equal the function itself.
    ops = {
        "autooc.operators": [
            "INITIALISATION",
            "SELECTION",
            "CROSSOVER",
            "MUTATION",
            "REPLACEMENT",
        ],
        "autooc.utilities.fitness": ["ERROR_METRIC"],
        "autooc.fitness": ["FITNESS_FUNCTION"],
        "autooc.algorithm": ["SEARCH_LOOP", "STEP"],
    }

    # We have to take 'algorithm' first as the functions from
    # algorithm need to be imported before any others to prevent
    # circular imports. We have to take 'utilities.fitness' before
    # 'fitness' because ERROR_METRIC has to be set in order to call
    # the fitness function constructor.

    for special_ops in ["autooc.algorithm", "autooc.utilities.fitness", "autooc.operators", "autooc.fitness"]:

        if all([callable(params[op]) for op in ops[special_ops]]):
            # params are already functions
            pass

        else:

            for op in ops[special_ops]:

                if special_ops == "autooc.fitness":
                    # Fitness functions represent a special case.

                    get_fit_func_imports()

                elif params[op] is not None:
                    # Split import name based on "." to find nested modules.
                    split_name = params[op].split(".")

                    if len(split_name) > 1:
                        # Check to see if a full path has been specified.

                        # Get attribute name.
                        attr_name = split_name[-1]

                        try:
                            # Try and use the exact specified path to load
                            # the module.

                            # Get module name.
                            module_name = ".".join(split_name[:-1])

                            # Import module and attribute and save.
                            params[op] = return_attr_from_module(
                                module_name, attr_name)

                        except Exception:
                            # Either a full path has not actually been
                            # specified, or the module doesn't exist. Try to
                            # append specified module to default location.

                            # Get module name.
                            module_name = ".".join(
                                [special_ops, ".".join(split_name[:-1])]
                            )

                            try:
                                # Import module and attribute and save.
                                params[op] = return_attr_from_module(
                                    module_name, attr_name
                                )

                            except Exception:
                                s = (
                                    "utilities.algorithm.initialise_run."
                                    "set_param_imports\n"
                                    "Error: Specified %s function not found:"
                                    " %s\n"
                                    "       Checked locations: %s\n"
                                    "                          %s\n"
                                    "       Please ensure parameter is "
                                    "specified correctly."
                                    % (
                                        op.lower(),
                                        attr_name,
                                        params[op],
                                        ".".join([module_name, attr_name]),
                                    )
                                )
                                raise Exception(s)

                    else:
                        # Just module name specified. Use default location.

                        # If multi-agent is specified need to change
                        # how search and step module is called
                        # Loop and step functions for multi-agent is contained
                        # inside algorithm search_loop_distributed and
                        # step_distributed respectively

                        if params["MULTIAGENT"] and (
                            op == "SEARCH_LOOP" or op == "STEP"
                        ):
                            # Define the directory structure for the multi-agent search
                            # loop and step
                            multiagent_ops = {
                                "search_loop": "distributed_algorithm.search_loop",
                                "step": "distributed_algorithm.step",
                            }

                            # Get module and attribute names
                            module_name = ".".join(
                                [special_ops, multiagent_ops[op.lower()]]
                            )
                            attr_name = split_name[-1]

                        else:
                            # Get module and attribute names.
                            module_name = ".".join([special_ops, op.lower()])
                            attr_name = split_name[-1]

                        # Import module and attribute and save.
                        print(module_name)
                        print(attr_name)
                        params[op] = return_attr_from_module(
                            module_name, attr_name)


def get_fit_func_imports():
    """
    Special handling needs to be done for fitness function imports,
    as fitness functions can be specified a number of different ways. Notably,
    a list of fitness functions can be specified, indicating multiple
    objective optimisation.

    Note that fitness functions must be classes where the class has the same
    name as its containing file. Fitness functions must be contained in the
    `fitness` module.

    :return: Nothing.
    """

    op = "FITNESS_FUNCTION"

    if "," in params[op]:
        # List of fitness functions given in parameters file.

        # Convert specified fitness functions into a list of strings.
        params[op] = params[op].strip("[()]").split(",")

    if isinstance(params[op], list) and len(params[op]) == 1:
        # Single fitness function given in a list format. Don't use
        # multi-objective optimisation.
        params[op] = params[op][0]

    if isinstance(params[op], list):
        # List of multiple fitness functions given.

        for i, name in enumerate(params[op]):

            # Split import name based on "." to find nested modules.
            split_name = name.strip().split(".")

            # Get module and attribute names.
            module_path = ".".join(["autooc.fitness", name.strip()])
            attr = split_name[-1]

            # Import this fitness function.
            params[op][i] = return_attr_from_module(module_path, attr)

        # Import base multi-objective fitness function class.
        from autooc.fitness.base_ff_classes.moo_ff import moo_ff

        # Set main fitness function as base multi-objective fitness
        # function class.
        params[op] = moo_ff(params[op])

    else:
        # A single fitness function has been specified.

        # Split import name based on "." to find nested modules.
        split_name = params[op].strip().split(".")

        # Get attribute name.
        attr_name = split_name[-1]

        # Get module name.
        module_name = ".".join(["autooc.fitness", params[op]])

        # Import module and attribute and save.
        params[op] = return_attr_from_module(module_name, attr_name)

        # Initialise fitness function.
        params[op] = params[op]()


def return_attr_from_module(module_name, attr_name):
    """
    Given a module path and the name of an attribute that exists in that
    module, import the attribute from the module using the importlib package
    and return it.

    :param module_name: The name/location of the desired module.
    :param attr_name: The name of the attribute.
    :return: The imported attribute from the module.
    """

    try:
        # Import module.
        module = importlib.import_module(module_name)

    except ModuleNotFoundError:
        s = (
            "utilities.algorithm.initialise_run.return_attr_from_module\n"
            "Error: Specified module not found: %s" % (module_name)
        )
        raise Exception(s)

    try:
        # Import specified attribute and return.
        return getattr(module, attr_name)

    except AttributeError:
        s = (
            "utilities.algorithm.initialise_run.return_attr_from_module\n"
            "Error: Specified attribute '%s' not found in module '%s'."
            % (attr_name, module_name)
        )
        raise Exception(s)


def pool_init(params_):
    """
    When initialising the pool the original params dict (params_) is passed in
    and used to update the newly created instance of params, as Windows does
    not retain the system memory of the parent process.

    :param params_: original params dict
    :return: Nothing.
    """

    from platform import system

    if system() == "Windows":
        params.update(params_)



from random import choice, randint, randrange

from autooc.algorithm.parameters import params
from autooc.representation.tree import Tree
from autooc.utilities.representation.check_methods import (get_nodes_and_depth,
                                                          ret_true)


def generate_tree(tree, genome, output, method, nodes, depth, max_depth, depth_limit):
    """
    Recursive function to derive a tree using a given method.

    :param tree: An instance of the Tree class.
    :param genome: The list of all codons in a tree.
    :param output: The list of all terminal nodes in a subtree. This is
    joined to become the phenotype.
    :param method: A string of the desired tree derivation method,
    e.g. "full" or "random".
    :param nodes: The total number of nodes in the tree.
    :param depth: The depth of the current node.
    :param max_depth: The maximum depth of any node in the tree.
    :param depth_limit: The maximum depth the tree can expand to.
    :return: genome, output, nodes, depth, max_depth.
    """

    # Increment nodes and depth, set depth of current node.
    nodes += 1
    depth += 1
    tree.depth = depth

    # Find the productions possible from the current root.
    productions = params["BNF_GRAMMAR"].rules[tree.root]

    if depth_limit:
        # Set remaining depth.
        remaining_depth = depth_limit - depth

    else:
        remaining_depth = depth_limit

    # Find which productions can be used based on the derivation method.
    available = legal_productions(
        method, remaining_depth, tree.root, productions["choices"]
    )

    # Randomly pick a production choice and make a codon with it.
    chosen_prod = choice(available)
    codon = generate_codon(chosen_prod, productions)

    # Set the codon for the current node and append codon to the genome.
    tree.codon = codon
    genome.append(codon)

    # Initialise empty list of children for current node.
    tree.children = []

    for symbol in chosen_prod["choice"]:
        # Iterate over all symbols in the chosen production.
        if symbol["type"] == "T":
            # The symbol is a terminal. Append new node to children.
            tree.children.append(Tree(symbol["symbol"], tree))

            # Append the terminal to the output list.
            output.append(symbol["symbol"])

        elif symbol["type"] == "NT":
            # The symbol is a non-terminal. Append new node to children.
            tree.children.append(Tree(symbol["symbol"], tree))

            # recurse on the new node.
            genome, output, nodes, d, max_depth = generate_tree(
                tree.children[-1],
                genome,
                output,
                method,
                nodes,
                depth,
                max_depth,
                depth_limit,
            )

    NT_kids = [
        kid for kid in tree.children if kid.root in params["BNF_GRAMMAR"].non_terminals
    ]

    if not NT_kids:
        # Then the branch terminates here
        depth += 1
        nodes += 1

    if depth > max_depth:
        # Set new maximum depth
        max_depth = depth

    return genome, output, nodes, depth, max_depth


def generate_codon(chosen_prod, productions):
    """
    Generate a single codon

    :param chosen_prod: the specific production to build a codon for
    :param productions: productions possible from the current root
    :return: a codon integer

    """

    # Find the index of the chosen production
    production_index = productions["choices"].index(chosen_prod)

    # Choose a random offset with guarantee that (offset + production_index) < codon_size
    offset = randrange(
        start=0,
        stop=params["BNF_GRAMMAR"].codon_size - productions["no_choices"] + 1,
        step=productions["no_choices"],
    )

    codon = offset + production_index
    return codon


def legal_productions(method, depth_limit, root, productions):
    """
    Returns the available production choices for a node given a specific
    depth limit.

    :param method: A string specifying the desired tree derivation method.
    Current methods are "random" or "full".
    :param depth_limit: The overall depth limit of the desired tree from the
    current node.
    :param root: The root of the current node.
    :param productions: The full list of production choices from the current
    root node.
    :return: The list of available production choices based on the specified
    derivation method.
    """

    # Get all information about root node
    root_info = params["BNF_GRAMMAR"].non_terminals[root]

    if method == "random":
        # Randomly build a tree.

        if depth_limit is None:
            # There is no depth limit, any production choice can be used.
            available = productions

        elif depth_limit > params["BNF_GRAMMAR"].max_arity + 1:
            # If the depth limit is greater than the maximum arity of the
            # grammar, then any production choice can be used.
            available = productions

        elif depth_limit < 0:
            # If we have already surpassed the depth limit, then list the
            # choices with the shortest terminating path.
            available = root_info["min_path"]

        else:
            # The depth limit is less than or equal to the maximum arity of
            # the grammar + 1. We have to be careful in selecting available
            # production choices lest we generate a tree which violates the
            # depth limit.
            available = [
                prod for prod in productions if prod["max_path"] <= depth_limit - 1
            ]

            if not available:
                # There are no available choices which do not violate the depth
                # limit. List the choices with the shortest terminating path.
                available = root_info["min_path"]

    elif method == "full":
        # Build a "full" tree where every branch extends to the depth limit.

        if depth_limit is None:
            # There is no depth limit specified for building a Full tree.
            # Raise an error as a depth limit HAS to be specified here.
            s = (
                "representation.derivation.legal_productions\n"
                "Error: Depth limit not specified for `Full` tree derivation."
            )
            raise Exception(s)

        elif depth_limit > params["BNF_GRAMMAR"].max_arity + 1:
            # If the depth limit is greater than the maximum arity of the
            # grammar, then only recursive production choices can be used.
            available = root_info["recursive"]

            if not available:
                # There are no recursive production choices for the current
                # rule. Pick any production choices.
                available = productions

        else:
            # The depth limit is less than or equal to the maximum arity of
            # the grammar + 1. We have to be careful in selecting available
            # production choices lest we generate a tree which violates the
            # depth limit.
            available = [
                prod for prod in productions if prod["max_path"] == depth_limit - 1
            ]

            if not available:
                # There are no available choices which extend exactly to the
                # depth limit. List the NT choices with the longest terminating
                # paths that don't violate the limit.
                available = [
                    prod for prod in productions if prod["max_path"] < depth_limit - 1
                ]

    return available


def pi_random_derivation(tree, max_depth):
    """
    Randomly builds a tree from a given root node up to a maximum given
    depth. Uses position independent methods to derive non-terminal nodes.
    Final tree is not guaranteed to reach the specified max_depth limit.

    :param tree: An instance of the representation.tree.Tree class.
    :param max_depth: The maximum depth to which to derive a tree.
    :return: The fully derived tree.
    """

    # Initialise derivation queue.
    queue = [
        [tree, ret_true(
            params["BNF_GRAMMAR"].non_terminals[tree.root]["recursive"])]
    ]

    # Initialise empty genome. With PI operators we can't use a depth-first
    # traversal of the tree to build the genome, we need to build it as we
    # encounter each node.
    genome = []

    while queue:
        # Loop until no items remain in the queue.

        # Pick a random item from the queue.
        chosen = randint(0, len(queue) - 1)

        # Pop the next item from the queue.
        all_node = queue.pop(chosen)
        node = all_node[0]

        # Get depth current node.
        if node.parent is not None:
            node.depth = node.parent.depth + 1

        # Find the productions possible from the current root.
        productions = params["BNF_GRAMMAR"].rules[node.root]

        # Set remaining depth.
        remaining_depth = max_depth - node.depth

        # Find which productions can be used based on the derivation method.
        available = legal_productions(
            "random", remaining_depth, node.root, productions["choices"]
        )

        # Randomly pick a production choice and make a codon with it.
        chosen_prod = choice(available)
        codon = generate_codon(chosen_prod, productions)

        # Set the codon for the current node and append codon to the genome.
        node.codon = codon

        # Insert codon into the genome.
        genome.append(codon)

        # Initialise empty list of children for current node.
        node.children = []

        for i, symbol in enumerate(chosen_prod["choice"]):
            # Iterate over all symbols in the chosen production.

            # Create new child.
            child = Tree(symbol["symbol"], node)

            # Append new node to children.
            node.children.append(child)

            if symbol["type"] == "NT":
                # The symbol is a non-terminal.

                # Check whether child is recursive
                recur_child = ret_true(
                    params["BNF_GRAMMAR"].non_terminals[child.root]["recursive"]
                )

                # Insert new child into the correct position in the queue.
                queue.insert(chosen + i, [child, recur_child])

    # genome, output, invalid, depth, and nodes can all be generated by
    # recursing through the tree once.
    _, output, invalid, depth, nodes = tree.get_tree_info(
        params["BNF_GRAMMAR"].non_terminals.keys(), [], []
    )

    return genome, output, nodes, depth


def pi_grow(tree, max_depth):
    """
    Grows a tree until a single branch reaches a specified depth. Does this
    by only using recursive production choices until a single branch of the
    tree has reached the specified maximum depth. After that any choices are
    allowed.

    :param tree: An instance of the representation.tree.Tree class.
    :param max_depth: The maximum depth to which to derive a tree.
    :return: The fully derived tree.
    """

    # Initialise derivation queue.
    queue = [
        [tree, ret_true(
            params["BNF_GRAMMAR"].non_terminals[tree.root]["recursive"])]
    ]

    # Initialise empty genome. With PI operators we can't use a depth-first
    # traversal of the tree to build the genome, we need to build it as we
    # encounter each node.
    genome = []

    while queue:
        # Loop until no items remain in the queue.

        # Pick a random item from the queue.
        chosen = randint(0, len(queue) - 1)

        # Pop the next item from the queue.
        all_node = queue.pop(chosen)
        node, recursive = all_node[0], all_node[0]

        # Get depth of current node.
        if node.parent is not None:
            node.depth = node.parent.depth + 1

        # Get maximum depth of overall tree.
        _, overall_depth = get_nodes_and_depth(tree)

        # Find the productions possible from the current root.
        productions = params["BNF_GRAMMAR"].rules[node.root]

        # Set remaining depth.
        remaining_depth = max_depth - node.depth

        if (overall_depth < max_depth) or (
            recursive and (not any([item[1] for item in queue]))
        ):
            # We want to prevent the tree from creating terminals until a
            # single branch has reached the full depth. Only select recursive
            # choices.

            # Find which productions can be used based on the derivation method.
            available = legal_productions(
                "full", remaining_depth, node.root, productions["choices"]
            )
        else:
            # Any production choices can be made.

            # Find which productions can be used based on the derivation method.
            available = legal_productions(
                "random", remaining_depth, node.root, productions["choices"]
            )

        # Randomly pick a production choice and make a codon with it.
        chosen_prod = choice(available)
        codon = generate_codon(chosen_prod, productions)

        # Set the codon for the current node and append codon to the genome.
        node.codon = codon

        # Insert codon into the genome.
        genome.append(codon)

        # Initialise empty list of children for current node.
        node.children = []

        for i, symbol in enumerate(chosen_prod["choice"]):
            # Iterate over all symbols in the chosen production.

            # Create new child.
            child = Tree(symbol["symbol"], node)

            # Append new node to children.
            node.children.append(child)

            if symbol["type"] == "NT":
                # The symbol is a non-terminal.

                # Check whether child is recursive
                recur_child = ret_true(
                    params["BNF_GRAMMAR"].non_terminals[child.root]["recursive"]
                )

                # Insert new child into the correct position in the queue.
                queue.insert(chosen + i, [child, recur_child])

    # genome, output, invalid, depth, and nodes can all be generated by
    # recursing through the tree once.
    _, output, invalid, depth, nodes = tree.get_tree_info(
        params["BNF_GRAMMAR"].non_terminals.keys(), [], []
    )

    return genome, output, nodes, depth



from autooc.algorithm.parameters import params
from autooc.stats.stats import stats


def clean_stats():
    """
    Removes certain unnecessary stats from the stats.stats.stats dictionary
    to clean up the current run.

    :return: Nothing.
    """

    if not params["CACHE"]:
        stats.pop("unique_inds")
        stats.pop("unused_search")

    if not params["MUTATE_DUPLICATES"]:
        try:
            stats.pop("regens")
        except Exception:
            print("Regens not found")



""" This program cues up and executes multiple runs of PYGE. Results of runs
    are parsed and placed in a spreadsheet for easy visual analysis.

    Copyright (c) 2014 Michael Fenton
    Hereby licensed under the GNU GPL v3."""
import sys
from multiprocessing import Pool
from subprocess import call
from sys import path

from autooc.algorithm.algorithm.general import check_python_version
from autooc.algorithm.parameters import params, set_params
from autooc.scripts.stats_parser import parse_stats_from_runs

path.append("../src")


check_python_version()


def execute_run(seed):
    """
    Initialise all aspects of a run.

    :return: Nothing.
    """

    exec_str = "python3 ponyge.py " "--random_seed " + str(seed) + " " + " ".join(
        sys.argv[1:]
    )

    call(exec_str, shell=True)


def execute_runs():
    """
    Execute multiple runs in series using multiple cores.

    :return: Nothing.
    """

    # Initialise empty list of results.
    results = []

    # Initialise pool of workers.
    pool = Pool(processes=params["CORES"])

    for run in range(params["RUNS"]):
        # Execute a single evolutionary run.
        results.append(pool.apply_async(execute_run, (run,)))

    for result in results:
        result.get()

    # Close pool once runs are finished.
    pool.close()


def check_params():
    """
    Checks the params to ensure an experiment name has been specified and
    that the number of runs has been specified.

    :return: Nothing.
    """

    if not params["EXPERIMENT_NAME"]:
        s = (
            "scripts.run_experiments.check_params\n"
            "Error: Experiment Name not specified.\n"
            "       Please specify a name for this set of runs."
        )
        raise Exception(s)

    if params["RUNS"] == 1:
        print("Warning: Only 1 run has been specified for this set of runs.")
        print(
            "         The number of runs can be specified with the command-"
            "line parameter `--runs`."
        )


def main():
    """
    The main function for running the experiment manager. Calls all functions.

    :return: Nothing.
    """

    # Setup run parameters.
    set_params(sys.argv[1:], create_files=False)

    # Check the correct parameters are set for this set of runs.
    check_params()

    # Execute multiple runs.
    execute_runs()

    # Save spreadsheets and all plots for all runs in the 'EXPERIMENT_NAME'
    # folder.
    parse_stats_from_runs(params["EXPERIMENT_NAME"])


if __name__ == "__main__":
    main()



from math import isnan

import numpy as np

np.seterr(all="raise")


class moo_ff:
    """
    Fitness function for multi-objective optimization problems. This fitness
    function acts as a holding class for all specified fitness functions.
    Individual fitness functions are held in an array called
    self.fitness_functions; when an individual is evaluated, it is evaluated
    on all fitness functions in this array.

    This is a holding class which exists just to be subclassed: it should not
    be instantiated.
    """

    # Required attribute for stats handling.
    multi_objective = True

    def __init__(self, fitness_functions):

        # Set list of individual fitness functions.
        self.fitness_functions = fitness_functions
        self.num_obj = len(fitness_functions)

        # Initialise individual fitness functions.
        for i, ff in enumerate(self.fitness_functions):
            self.fitness_functions[i] = ff()

        # Set up list of default fitness values (as per individual fitness
        # functions).
        self.default_fitness = []
        for f in fitness_functions:
            self.default_fitness.append(f.default_fitness)

    def __call__(self, ind):
        """
        Note that math functions used in the solutions are imported from either
        utilities.fitness.math_functions or called from numpy.

        :param ind: An individual to be evaluated.
        :return: The fitness of the evaluated individual.
        """

        # the multi-objective fitness is defined as a list of values, each one
        # representing the output of one objective function. The computation is
        # made by the function multi_objc_eval, implemented by a subclass,
        # according to the problem.
        fitness = [ff(ind) for ff in self.fitness_functions]

        if any([isnan(i) for i in fitness]):
            # Check if any objective fitness value is NaN, if so set default
            # fitness.
            fitness = self.default_fitness

        return fitness

    @staticmethod
    def value(fitness_vector, objective_index):
        """
        This is a static method required by NSGA-II for sorting populations
        based on a given fitness function, or for returning a given index of a
        population based on a given fitness function.

        :param fitness_vector: A vector/list of fitnesses.
        :param objective_index: The index of the desired fitness.
        :return: The fitness at the objective index of the fitness vector.
        """

        if not isinstance(fitness_vector, list):
            return float("inf")

        return fitness_vector[objective_index]



import dtw  # https://pypi.python.org/pypi/dtw
import editdistance  # https://pypi.python.org/pypi/editdistance
import lzstring  # https://pypi.python.org/pypi/lzstring/

from autooc.algorithm.parameters import params
from autooc.fitness.base_ff_classes.base_ff import base_ff

"""

This fitness function is for a sequence-match problem: we're given
an integer sequence target, say [0, 5, 0, 5, 0, 5], and we try to synthesize a
program (loops, if-statements, etc) which will *yield* that sequence,
one item at a time.

There are several components of the fitness:

1. concerning the program:
    i. length of the program (shorter is better)
    ii. compressibility of the program (non-compressible, ie DRY, is better)

2. concerning distance from the target:
    i. dynamic time warping distance from the program's output to the target
    (lower is better).
    ii. Levenshtein distance from the program's output to the target
    (lower is better).

"""


# available for use in synthesized programs
def succ(n, maxv=6):
    """
    Available for use in synthesized programs.

    :param n:
    :param maxv:
    :return:
    """

    return min(n + 1, maxv)


def pred(n, minv=0):
    """
    Available for use in synthesized programs.

    :param n:
    :param minv:
    :return:
    """

    return max(n - 1, minv)


def truncate(n, g):
    """
    the program will yield one item at a time, potentially forever. We only
    up to n items.

    :param n:
    :param g:
    :return:
    """

    for i in range(n):
        yield next(g)


def dist(t0, x0):
    """
    numerical difference, used as a component in DTW.

    :param t0:
    :param x0:
    :return:
    """

    return abs(t0 - x0)


def dtw_dist(s, t):
    """
    Dynamic time warping distance between two sequences.

    :param s:
    :param t:
    :return:
    """

    s = list(map(int, s))
    t = list(map(int, t))
    d, M, C, path = dtw.dtw(s, t, dist)

    return d


def lev_dist(s, t):
    """
    Levenshtein distance between two sequences, normalised by length of the
    target -- hence this is *asymmetric*, not really a distance. Don't
    normalise by length of the longer, because it would encourage evolution
    to create longer and longer sequences.

    :param s:
    :param t:
    :return:
    """

    return editdistance.eval(s, t) / len(s)


def compress(s):
    """
    Convert to a string and compress. lzstring is a special-purpose compressor,
    more suitable for short strings than typical compressors.

    :param s:
    :return:
    """

    s = "".join(map(str, s))
    return lzstring.LZString().compress(s)


def compressibility(s):
    """
    Compressibility is in [0, 1]. It's high when the compressed string
    is much shorter than the original.

    :param s:
    :return:
    """

    return 1 - len(compress(s)) / len(s)


def proglen(s):
    """
    Program length is measured in characters, but in order to keep the values
    in a similar range to that of compressibility, DTW and Levenshtein, we
    divide by 100. This is a bit arbitrary.

    :param s: A string of a program phenotype.
    :return: The length of the program divided by 100.
    """

    return len(s) / 100.0


class sequence_match(base_ff):
    def __init__(self):
        """
        Initilise class instance
        """
        # Initialise base fitness function class.
        super().__init__()

        # --target will be a sequence such as (0, 5, 0, 5)
        self.target = eval(params["TARGET"])

        # we assume --extra_parameters is a comma-separated kv sequence, eg:
        # "alpha=0.5, beta=0.5, gamma=0.5"
        # which we can pass to the dict() constructor
        extra_fit_params = eval("dict(" + params["EXTRA_PARAMETERS"] + ")")
        self.alpha = extra_fit_params["alpha"]
        self.beta = extra_fit_params["beta"]
        self.gamma = extra_fit_params["gamma"]

    def evaluate(self, ind, **kwargs):
        """
        ind.phenotype will be a string incl fn defns etc. when we exec it
        will create a value XXX_output_XXX, but we exec inside an empty dict
        for safety. But we put a couple of useful primitives in the dict too.

        :param ind:
        :return:
        """

        p, d = ind.phenotype, {"pred": pred, "succ": succ}
        exec(p, d)

        # this is the program's output: a generator
        s = d["XXX_output_XXX"]

        # Truncate the generator and convert to list
        s = list(truncate(len(self.target), s))

        # Set target
        t = self.target

        # various weightings of four aspects of our fitness. the formula is:
        # fitness = gamma * dist + (1 - gamma) * length
        # where dist = alpha * lev_dist(t, s) + (1 - alpha) * dtw_dist(t, s)
        # and length = beta * proglen(t) + (1 - beta) * compressibility(t)
        # but when any of alpha, beta and gamma is 0 or 1, we can save some
        # calculation:

        if self.gamma > 0.0:
            if self.alpha > 0.0:
                lev_dist_v = lev_dist(t, s)
            else:
                lev_dist_v = 0.0
            if self.alpha < 1.0:
                dtw_dist_v = dtw_dist(t, s)
            else:
                dtw_dist_v = 0.0
            dist_v = self.alpha * lev_dist_v + (1 - self.alpha) * dtw_dist_v
        else:
            dist_v = 0.0

        if self.gamma < 1.0:
            if self.beta > 0.0:
                proglen_v = proglen(p)
            else:
                proglen_v = 0.0
            if self.beta < 1.0:
                compressibility_v = compressibility(p)
            else:
                compressibility_v = 0.0
            length_v = self.beta * proglen_v + \
                (1 - self.beta) * compressibility_v
        else:
            length_v = 0.0

        return self.gamma * dist_v + (1 - self.gamma) * length_v


if __name__ == "__main__":
    # TODO write some tests here
    pass



from autooc.algorithm.parameters import params


class Tree:
    def __init__(self, expr, parent):
        """
        Initialise an instance of the tree class.

        :param expr: A non-terminal from the params['BNF_GRAMMAR'].
        :param parent: The parent of the current node. None if node is tree
        root.
        """

        self.parent = parent
        self.codon = None
        self.depth = 1
        self.root = expr
        self.children = []
        self.snippet = None

    def __str__(self):
        """
        Builds a string of the current tree.

        :return: A string of the current tree.
        """

        # Initialise the output string.
        result = "("

        # Append the root of the current node to the output string.
        result += str(self.root)

        for child in self.children:
            # Iterate across all children.

            if len(child.children) > 0:
                # Recurse through all children.
                result += " " + str(child)

            else:
                # Child is a terminal, append root to string.
                result += " " + str(child.root)

        result += ")"

        return result

    def __copy__(self):
        """
        Creates a new unique copy of self.

        :return: A new unique copy of self.
        """

        # Copy current tree by initialising a new instance of the tree class.
        tree_copy = Tree(self.root, self.parent)

        # Set node parameters.
        tree_copy.codon, tree_copy.depth = self.codon, self.depth

        tree_copy.snippet = self.snippet

        for child in self.children:
            # Recurse through all children.
            new_child = child.__copy__()

            # Set the parent of the copied child as the copied parent.
            new_child.parent = tree_copy

            # Append the copied child to the copied parent.
            tree_copy.children.append(new_child)

        return tree_copy

    def __eq__(self, other, same=True):
        """
        Set the definition for comparison of two instances of the tree
        class by their attributes. Returns True if self == other.

        :param other: Another instance of the tree class with which to compare.
        :return: True if self == other.
        """

        # Get attributes of self and other.
        a_self, a_other = vars(self), vars(other)

        # Don't look at the children as they are class instances themselves.
        taboo = ["parent", "children", "snippet", "id"]
        self_no_kids = {k: v for k, v in a_self.items() if k not in taboo}
        other_no_kids = {k: v for k, v in a_other.items() if k not in taboo}

        # Compare attributes
        if self_no_kids != other_no_kids:
            # Attributes are not the same.
            return False

        else:
            # Attributes are the same
            child_list = [self.children, other.children]

            if len(list(filter(lambda x: x is not None, child_list))) % 2 != 0:
                # One contains children, the other doesn't.
                return False

            elif self.children and len(self.children) != len(other.children):
                # Number of children differs between self and other.
                return False

            elif self.children:
                # Compare children recursively.
                for i, child in enumerate(self.children):
                    same = child.__eq__(other.children[i], same)

        return same

    def get_target_nodes(self, array, target=None):
        """
        Returns the all NT nodes which match the target NT list in a
        given tree.

        :param array: The array of all nodes that match the target.
        :param target: The target nodes to match.
        :return: The array of all nodes that match the target.
        """

        if self.root in target:
            # Check if the current node matches the target.

            # Add the current node to the array.
            array.append(self)

        # Find all non-terminal children of the current node.
        NT_kids = [
            kid
            for kid in self.children
            if kid.root in params["BNF_GRAMMAR"].non_terminals
        ]

        for child in NT_kids:
            if NT_kids:
                # Recursively call function on any non-terminal children.
                array = child.get_target_nodes(array, target=target)

        return array

    def get_node_labels(self, labels):
        """
        Recurses through a tree and appends all node roots to a set.

        :param labels: The set of roots of all nodes in the tree.
        :return: The set of roots of all nodes in the tree.
        """

        # Add the current root to the set of all labels.
        labels.add(self.root)

        for child in self.children:
            # Recurse on all children.
            labels = child.get_node_labels(labels)

        return labels

    def get_tree_info(
        self, nt_keys, genome, output, invalid=False, max_depth=0, nodes=0
    ):
        """
        Recurses through a tree and returns all necessary information on a
        tree required to generate an individual.

        :param genome: The list of all codons in a subtree.
        :param output: The list of all terminal nodes in a subtree. This is
        joined to become the phenotype.
        :param invalid: A boolean flag for whether a tree is fully expanded.
        True if invalid (unexpanded).
        :param nt_keys: The list of all non-terminals in the grammar.
        :param nodes: the number of nodes in a tree.
        :param max_depth: The maximum depth of any node in the tree.
        :return: genome, output, invalid, max_depth, nodes.
        """

        # Increment number of nodes in tree and set current node id.
        nodes += 1

        if self.parent:
            # If current node has a parent, increment current depth from
            # parent depth.
            self.depth = self.parent.depth + 1

        else:
            # Current node is tree root, set depth to 1.
            self.depth = 1

        if self.depth > max_depth:
            # Set new max tree depth.
            max_depth = self.depth

        if self.codon:
            # If the current node has a codon, append it to the genome.
            genome.append(self.codon)

        # Find all non-terminal children of current node.
        NT_children = [
            child for child in self.children if child.root in nt_keys]

        if not NT_children:
            # The current node has only terminal children, increment number
            # of tree nodes.
            nodes += 1

            # Terminal children increase the current node depth by one.
            # Check the recorded max_depth.
            if self.depth + 1 > max_depth:
                # Set new max tree depth.
                max_depth = self.depth + 1

        if self.root in nt_keys and not self.children:
            # Current NT has no children. Invalid tree.
            invalid = True

        for child in self.children:
            # Recurse on all children.

            if not child.children:
                # If the current child has no children it is a terminal.
                # Append it to the phenotype output.
                output.append(child.root)

                if child.root in nt_keys:
                    # Current non-terminal node has no children; invalid tree.
                    invalid = True

            else:
                # The current child has children, recurse.
                genome, output, invalid, max_depth, nodes = child.get_tree_info(
                    nt_keys, genome, output, invalid, max_depth, nodes
                )

        return genome, output, invalid, max_depth, nodes

    def print_tree(self):
        """
        Prints out all nodes in the tree, indented according to node depth.

        :return: Nothing.
        """

        print(self.depth, "".join([" " for _ in range(self.depth)]), self.root)

        for child in self.children:
            if not child.children:
                print(
                    self.depth + 1,
                    "".join([" " for _ in range(self.depth + 1)]),
                    child.root,
                )
            else:
                child.print_tree()
